from ._base import MdElement


class MdQuote(MdElement):
    """Represents a Markdown blockquote.

    Attributes:
        content (str or MdElement): The content of the quote.

    Examples:
        >>> from mdfy.elements import MdQuote
        >>>
        >>> quote = MdQuote("This is a quote.")
        >>> print(quote)
        > This is a quote.
        >>>
        >>> quote = MdQuote("This is a quote.\\nThis is another line.")
        >>> print(quote)
        > This is a quote.
        > This is another line.
    """

    def __init__(self, content: str) -> None:
        """Initializes an instance of the MdQuote class to represent a Markdown blockquote.

        Args:
            content (str or MdElement): The content of the quote.
        """
        self.content = content

    def __str__(self) -> str:
        """Returns a string representation of the blockquote in Markdown format.

        Returns:
            str: String representation of the blockquote.
        """
        # Convert the content to string if it's an instance of MdElement
        content_str = (
            str(self.content) if isinstance(self.content, MdElement) else self.content
        )
        content_str = "" if content_str is None else content_str

        # Ensure content is not None
        lines = [""] if len(content_str) <= 0 else content_str.splitlines()

        # Create a blockquote by prefixing each line with '>'
        return "\n".join([f"> {line}" for line in lines])



from ._base import MdElement
from .code import MdCode
from .header import MdHeader
from .horizontal import MdHorizontal
from .image import MdImage
from .link import MdLink
from .list import MdList
from .quote import MdQuote
from .table import MdTable
from .text import MdText



import logging

from ._base import MdElement

logger = logging.getLogger(__name__)


class MdCode(MdElement):
    """Represents Markdown code or code block.

    Attributes:
        code (str): The code string.
        inline (bool): Determines if the code should be represented as inline or block.
        syntax (str): The programming language syntax for the code block.

    Examples:
        >>> code = MdCode("print('Hello World!')", inline=True)
        >>> print(code)
        `print('Hello World!')`
        >>> code = MdCode("print('Hello World!')", inline=False, syntax="python")
        >>> print(code)
        ```python
        print('Hello World!')
        ```
    """

    def __init__(self, code: str, inline: bool = False, syntax: str = "") -> None:
        """Initializes an instance of the MdCode class to represent Markdown code or code block.

        Args:
            code (str): The code string.
            inline (bool, optional): Determines if the code should be represented as inline or block.
                                      Defaults to block (False).
            syntax (str, optional): The programming language syntax for the code block.
                                    This only has an effect when inline is False.
        """

        if code is None:
            logger.warning("Code content is None, setting to empty string")
            code = ""

        self.code = code
        self.inline = inline
        self.syntax = syntax

        # Check for newlines to override inline setting
        if "\n" in code:
            self.inline = False

    def __str__(self) -> str:
        """Returns a string representation of the code in Markdown format.

        Returns:
            str: String representation of the code.

        Warnings:
            If the code is None, it will log a warning and set the code to an empty string.
        """

        code = self.code
        if self.inline:
            return f"`{code}`"
        else:
            return f"```{self.syntax}\n{code}\n```"



import warnings

from ._base import MdElement


class MdHorizontal(MdElement):
    """Represents a Markdown horizontal rule.

    Attributes:
        content (str): The content for representing the horizontal rule.

    Examples:
        >>> from mdfy.elements import MdHorizontal
        >>>
        >>> horizontal = MdHorizontal()
        >>> print(horizontal)
        ***
        >>>
        >>> horizontal = MdHorizontal("---")
        >>> print(horizontal)
        ---
    """

    def __init__(self, content: str = "***") -> None:
        """Initializes an instance of the MdHorizontal class to represent a Markdown horizontal rule.

        Args:
            content (str, optional): The content of the horizontal rule. Defaults to "***".
        """
        if not isinstance(content, str):
            warnings.warn(
                f"Horizontal content is not a string, converting to string: {content}"
            )
            content = str(content)
        self.content = content

    def __str__(self) -> str:
        return f"\n{self.content}\n"



import logging

from ._base import MdElement

logger = logging.getLogger(__name__)


class MdImage(MdElement):
    """Represents a Markdown image.

    Attributes:
        src (str): The source URL or path of the image.
        alt (str): The alternative text for the image.

    Examples:
        >>> from mdfy.elements import MdImage
        >>>
        >>> image = MdImage("https://example.com/image.png")
        >>> print(image)
        ![](https://example.com/image.png)
        >>>
        >>> image = MdImage("https://example.com/image.png", alt="Example image")
        >>> print(image)
        ![Example image](https://example.com/image.png)
    """

    def __init__(self, src: str, alt: str = "") -> None:
        """Initializes an instance of the MdImage class to represent a Markdown image.

        Args:
            src (str): The source URL or path of the image.
            alt (str, optional): The alternative text for the image. Defaults to an empty string.
        """
        self.src = src
        self.alt = alt

    def __str__(self) -> str:
        """Returns a string representation of the image in Markdown format.

        Returns:
            str: String representation of the image.

        Warnings:
            If the image source is None, it will log a warning and set the source to an empty string.
        """
        src = self.src
        if src is None:
            logger.warning("Image source is None, setting to empty string")
            src = ""
        return f"![{self.alt}]({src})"



import warnings
from typing import Optional

from ._base import MdElement


class MdLink(MdElement):
    """Represents a Markdown link.

    Attributes:
        url (str): The URL of the link.
        text (str): The display text for the link.
        title (str): The title attribute for the link.

    Examples:
        >>> from mdfy.elements import MdLink
        >>>
        >>> link = MdLink("https://www.example.com", "example")
        >>> print(link)
        [example](https://www.example.com)
        >>>
        >>> link = MdLink("https://www.example.com", "example", title="example")
        >>> print(link)
        [example](https://www.example.com "example")
    """

    def __init__(self, url: str, text: str = "", title: Optional[str] = None) -> None:
        """Initializes an instance of the MdLink class to represent a Markdown link.

        Args:
            url (str): The URL of the link.
            text (str, optional): The display text for the link. Defaults to an empty string.
            title (str, optional): The title attribute for the link. Defaults to None.
        """
        self.url = url
        self.text = text
        self.title = title

    def __str__(self) -> str:
        """Returns a string representation of the link in Markdown format.

        Returns:
            str: String representation of the link.

        Warnings:
            If the link URL is None, it will log a warning and set the URL to an empty string.
        """
        url = self.url
        if url is None:
            warnings.warn("Link URL is None, setting to empty string")
            url = ""
        text_str = f"{self.text}" if self.text else url
        title_str = f' "{self.title}"' if self.title else ""
        return f"[{text_str}]({url}{title_str})"



from pathlib import Path
from typing import List, Union

from .elements import MdElement


def write(content: str, filepath: str) -> None:
    with open(filepath, "w") as file:
        for item in content:
            file.write(str(item) + "\n")


class Mdfier:
    """Writes Markdown content to a file.

    Attributes:
        filepath (Path): The path to the file.

    Examples:
        >>> from mdfy import Mdfier
        >>> from mdfy.elements import MdHeader, MdQuote
        >>>
        >>> mdfier = Mdfier("README.md")
        >>> mdfier.write([
        ...     MdHeader("Hello, world!", 1),
        ...     MdQuote("This is a quote.")
        ... ])
        >>>
        >>> with open("README.md") as file:
        ...     print(file.read())
        ...
        # Hello, world!
        > This is a quote.
    """

    def __init__(self, filepath: Union[str, Path]) -> None:
        """Initializes an instance of the Mdfier class to write Markdown content to a file.

        Args:
            filepath (Union[str, Path]): The path to the file.
        """

        self.filepath = Path(filepath)
        self.filepath.parent.mkdir(parents=True, exist_ok=True)

    @classmethod
    def stringify(cls, contents: List[Union[str, MdElement]]) -> str:
        """Converts the given Markdown content to a string.

        Args:
            content (Union[str, MdElement]): The Markdown content to convert to a string.
        """

        return "\n".join([str(item) for item in contents])

    def write(self, contents: List[Union[str, MdElement]]) -> None:
        """Writes the given Markdown content to the file.

        Args:
            content (Union[str, MdElement]): The Markdown content to write to the file.
        """

        markdown = self.stringify(contents)
        with open(self.filepath, "w") as file:
            file.write(markdown + "\n")



import abc
import re
from typing import Dict, Optional

from lark import Lark, Token, Tree
from lark.visitors import Interpreter

from ._base import MdElement


class MdTextInterpreter(Interpreter):
    """Lark Interpreter tailored for MdText interpreting styled text.

    Args:
        style_patterns (Dict[str, str]): A dictionary mapping style names to formatting strings.
    """

    def __init__(self, style_patterns: Dict[str, str]):
        super().__init__()
        self.style_patterns = style_patterns

    def process_content_tree(self, tree: Tree) -> str:
        """Processes a subtree for content in yled text.

        Args:
            tree (Tree): The subtree to process.

        Returns:
            str: The processed text.
        """
        interpreted = ""
        for child in tree.children:
            if isinstance(child, Token):
                interpreted += child.value
            elif isinstance(child, Tree):
                interpreted += self.styled_text(child)
            else:
                raise ValueError(f"Unable to handle this child type: {child.__class__}")
        return interpreted

    def styled_text(self, tree: Tree) -> str:
        """Processes a subtree for styled text.

        Args:
            tree (Tree): The subtree to process.

        Returns:
            str: The styled text.
        """
        _, *content, _, style, _ = tree.children

        target_text = ""
        for tree in content:
            target_text += self.process_content_tree(tree)

        if not isinstance(style, Token):
            raise ValueError(f"Expected style to be a Token, got {style.__class__}")

        if style.value in self.style_patterns:
            return self.style_patterns[style.value].format(target_text)
        else:
            return target_text

    def start(self, tree: Tree) -> str:
        return "".join(self.visit_children(tree))


grammar = r"""
    start: TEXT? styled_text? (start)* TEXT?
    styled_text: LBRACE content+ COLON INTEXT RBRACE
    content: INTEXT | INTEXT? styled_text
    TEXT:   WS* /[^{}]+/ WS*
    INTEXT: WS* /[^{}:]+/ WS*
    LBRACE: "{"
    RBRACE: "}"
    COLON: ":"

    %import common.WS
"""


class MdFormatter(abc.ABC):
    """Abstract base class for Markdown formatters."""

    @abc.abstractmethod
    def format(self, text: str) -> str:
        """Formats the text and returns the formatted text.

        Args:
            text (str): The text to be formatted.

        Returns:
            str: The formatted text.
        """
        raise NotImplementedError


class MdTextFormatter(MdFormatter):
    """Markdown Text Formatter to handle text styling based on a specified grammar and style patterns.

    Attributes:
        grammar (str): The grammar to be used for parsing the text.
        interpreter (Interpreter): The interpreter to be used for interpreting the parsed text.
        patterns (Dict[str, str]): Expanded style patterns including aliases.
    """

    STYLE_PATTERNS = {
        "strong": "***{}***",
        "bold": "**{}**",
        "italic": "*{}*",
        "not": "~~{}~~",
        "underline": "<u>{}</u>",
        "quote": "`{}`",
    }

    STYLE_ALIASES = {
        "strong": ["st"],
        "bold": ["bo", "bd"],
        "italic": ["it"],
        "not": ["no", "nt"],
        "underline": ["un", "ul"],
        "quote": ["qu", "qt"],
    }

    def __init__(
        self,
        grammar: str = grammar,
        interpreter: Optional[Interpreter] = None,
        patterns: Optional[dict] = None,
    ):
        self.grammar = grammar
        if self.grammar is None:
            raise ValueError("Grammar cannot be None")
        self.parser = Lark(self.grammar)

        if patterns is None:
            patterns = self.expand_style_patterns(
                self.STYLE_PATTERNS, self.STYLE_ALIASES
            )
        self.patterns = patterns

        if interpreter is None:
            interpreter = MdTextInterpreter(self.patterns)
        self.interpreter = interpreter

    def expand_style_patterns(self, base_patterns: dict, aliases: dict) -> dict:
        """Expands the style patterns with aliases.

        Args:
            base_patterns (Dict[str, str]): Base style patterns.
            aliases (Dict[str, str]): Aliases for the style names.

        Returns:
            Dict[str, str]: Expanded style patterns.
        """
        expand_patterns = {
            alias: base_patterns[style_name]
            for style_name, aliases in aliases.items()
            for alias in aliases
        }
        return {**base_patterns, **expand_patterns}

    def format(self, text: str) -> str:
        """Formats the text for style markers and returns the formatted text.

        Args:
            text (str): The text to be formatted.

        Returns:
            str: The formatted text.
        """
        parsed = self.parse(text)
        result: str = self.interpreter.visit(parsed)
        return result

    def parse(self, text: str) -> Tree:
        """Parses the text for style markers and returns the parsed text.

        Args:
            text (str): The text to be parsed.

        Returns:
            lark.Tree: The parsed syntax tree.
        """
        return self.parser.parse(text)


class MdText(MdElement):
    """MdElementt class to handle the text and styling of text.

    Attributes:
        content (str): The content string containing potential style markers.
        formatter (MdFormatter): The formatter to apply styling to the content.

    Examples:
        >>> text = MdText("This is {bold:bold} text.")
        >>> print(text)
        This is **bold** text.
        >>> MdText("{This is underline text:underline}.").to_str()
        <u>This is underline text</u>.
        >>> MdText("{This is {nested:bold} style text:underline}.").to_str()
        <u>This is **nested** style text</u>.
        >>> MdText("You can use aliases e.g. {st:st}  {bd:bo}.").to_str()
        You can use aliases e.g. ***st***  **bd**.

    Note:
        Available style patterns:
            - strong: bold text (e.g. `***strong***`)
            - bold: bold text (e.g. `**bold**`)
            - italic: italic text (e.g. {*italic*})
            - not: strike-through text (e.g. `~~strike-through~~`)
            - underline: underlined text (e.g. `<u>underlined</u>`)
            - quote: quoted text (e.g. `\`quoted\``)

        Also, the following aliases are available for the style patterns:
            - strong: st
            - bold: bo, bd
            - italic: it
            - not: no, nt
            - underline: un, ul
            - quote: qu, qt
    """

    def __init__(self, content: str, formatter: MdFormatter = MdTextFormatter()):
        """Initializes an instance of the MdText class to handle the text and styling of text.

        Args:
            content (str): The content string containing potential style markers.
            formatter (MdFormatter, optional): The formatter to apply styling to the content.
                                                     Defaults to MdTextFormatter().
        """

        self.content = content
        self.formatter = formatter

    def __str__(self) -> str:
        """Returns the styled content as per the specified style markers.

        Returns:
            str: Formatted markdown string with the appropriate styles applied.
        """
        result = self.content
        result = self.formatter.format(result)

        return result

    def __add__(self, other: "MdText") -> "MdText":
        """Adds two MdText objects together.

        Args:
            other (MdText): The other MdText object to be added.

        Returns:
            MdText: A new MdText object containing the concatenated content of the two objects.
        """
        return MdText(str(self) + str(other))



import warnings
from typing import Any, Dict, List, Optional, Union

from ._base import MdElement


class MdTable(MdElement):
    """Converter for dict or list to markdown table.

    Args:
        data (dict or list): The data to convert.
        transpose (bool, optional): If True, transpose the table. Defaults to False.
        labels (list[str], optional): Label for header when transposed. Defaults to 'Key' and 'value {i}'.
        precision (Optional[int]): Number of decimal places for floats. If None, values are not formatted.

    Examples:
        >>> data = {
        ...     "Name": "John Doe",
        ...     "Age": 30,
        ...     "Occupation": "Software Engineer",
        ... }
        >>> table = MdTable(data)
        >>> print(table)
        | Name | Age | Occupation |
        | --- | --- | --- |
        | John Doe | 30 | Software Engineer |
        >>> table = MdTable(data, transpose=True)
        >>> print(table)
        | Key | Value 0 | Value 1 | Value 2 |
        | --- | --- | --- | --- |
        | Name | John Doe |  |  |
        | Age | 30 |  |  |
        | Occupation | Software Engineer |  |  |
        >>> table = MdTable(data, transpose=True, labels=["Header", "Value"])
        >>> print(table)
        | Header | Value |
        | --- | --- |
        | Name | John Doe |
        | Age | 30 |
        | Occupation | Software Engineer |
        >>> # you can use nested dict
        >>> data = {
        ...     "Name": "John Doe",
        ...     "Age": 30,
        ...     "Occupation": "Software Engineer",
        ...     "Address": {
        ...         "Street": "123 Main St",
        ...         "City": "Anytown",
        ...         "State": "CA",
        ...         "Zip": 12345,
        ...     },
        ... }
        >>> table = MdTable(data)
        >>> print(table)
        | Name | Age | Occupation | Address.Street | Address.City | Address.State | Address.Zip |
        | --- | --- | --- | --- | --- | --- | --- |
        | John Doe | 30 | Software Engineer | 123 Main St | Anytown | CA | 12345 |
        >>> # you can also use a list of dicts
        >>> data = [
        ...     {
        ...         "Name": "John Doe",
        ...         "Age": 30,
        ...         "Occupation": "Software Engineer",
        ...     },
        ...     {
        ...         "Name": "Jane Doe",
        ...         "Age": 25,
        ...         "Occupation": "Data Scientist",
        ...     },
        ... ]
        >>> table = MdTable(data)
        >>> print(table)
        | Name | Age | Occupation |
        | --- | --- | --- |
        | John Doe | 30 | Software Engineer |
        | Jane Doe | 25 | Data Scientist |
        >>> # you can specify the precision for floats
        >>> data = {
        ...     "Name": "John Doe",
        ...     "Age": 30,
        ...     "Occupation": "Software Engineer",
        ...     "Height": 1.83,
        ...     "Weight": 80.5,
        ... }
        >>> table = MdTable(data, precision=2)
        >>> print(table)
        | Name | Age | Occupation | Height | Weight |
        | --- | --- | --- | --- | --- |
        | John Doe | 30 | Software Engineer | 1.83 | 80.50 |
    """

    def __init__(
        self,
        data: Union[Dict[str, Any], List[Dict[str, Any]]],
        transpose: bool = False,
        labels: Optional[List[str]] = None,
        precision: Union[None, int] = None,
    ):
        """Initialize a MdTable instance.

        Args:
            data (Union[Dict[str, Any], List[Dict[str, Any]]]): The data to convert.
            transpose (bool, optional): If True, transpose the table. Defaults to False.
            labels (list[str], optional): Label for header when transposed. Defaults to 'Key' and 'value {i}'.
            precision (Optional[int]): Number of decimal places for floats. If None, values are not formatted.
        """

        if isinstance(data, list):
            pass
        elif isinstance(data, dict):
            data = [data]
        else:
            raise ValueError(
                "Provided data is not a dictionary or list of dictionaries"
            )
        self.data = self._flatten_data(data)

        self.transpose = transpose
        self.precision = precision
        self.labels = labels

    def _flatten_data(self, data: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        """Flatten nested dictionaries in the data.

        Args:
            data (List[Dict[str, Any]]): The data to flatten.
        """
        flattened_data = []
        for entry in data:
            flattened_data.append(self._flatten_dict(entry))

        return flattened_data

    def _flatten_dict(
        self, d: Dict, parent_key: str = "", sep: str = "."
    ) -> Dict[str, Any]:
        """Recursively flatten a nested dictionary.

        Args:
            d (Dict): Dictionary to flatten.
            parent_key (str, optional): Key from parent dictionary. Defaults to ''.
            sep (str, optional): Separator to use between keys. Defaults to '.'.

        Returns:
            Dict: Flattened dictionary.
        """
        items = {}
        for k, v in d.items():
            new_key = f"{parent_key}{sep}{k}" if parent_key else k
            if isinstance(v, dict):
                items.update(self._flatten_dict(v, new_key, sep=sep))
            else:
                items[new_key] = v
        return items

    def dict_to_md_table(
        self,
        transpose: bool = False,
        labels: Optional[List[str]] = None,
        output_file: Union[None, str] = None,
        precision: Union[None, int] = None,
    ) -> str:
        """Convert the data to a Markdown formatted table.

        Args:
            transpose (bool, optional): If True, transpose the table. Defaults to False.
            labels (list[str], optional): Label for header when transposed. Defaults to 'Key' and 'value {i}'.
            output_file (Union[None, str], optional): The filename to write the table to.
            precision (Union[None, int], optional): The precision for floating point numbers.

        Returns:
            str: Markdown formatted table.
        """

        self.precision = precision

        md_table = ""
        if not self.data:
            return md_table

        if transpose:
            md_table = self._create_transposed_table(labels=self.labels)
        else:
            md_table = self._create_regular_table()

        if output_file:
            with open(output_file, "w") as f:
                f.write(md_table)

        return md_table

    def _create_transposed_table(self, labels: Optional[List[str]]) -> str:
        """Create a transposed version of the table.

        Args:
            key_label (str): Label for keys.
            value_label (str): Label for values.

        Returns:
            str: Markdown formatted transposed table.
        """
        transposed_data = self._transpose_data()
        if labels is None:
            labels = ["Key"] + [f"Value {i}" for i in range(len(self.data))]
        headers = self._create_transposed_headers(labels)
        table_rows = self._create_table_rows(transposed_data)

        return headers + "\n" + table_rows

    def _transpose_data(self) -> Dict[str, List[Any]]:
        """Transpose the data for creating a transposed table.

        Returns:
            dict: Transposed data as a dictionary.
        """
        headers = self.data[0].keys()
        transposed_data: Dict[str, list] = {header: [] for header in headers}
        for row in self.data:
            for key, value in row.items():
                transposed_data[key].append(value)

        return transposed_data

    def _create_transposed_headers(self, labels: List[str]) -> str:
        """Create headers for the transposed table.

        Args:
            key_label (str): Label for key.
            value_label (str): Label for values.

        Returns:
            str: Transposed table headers.
        """
        num_columns = len(self.data) + 1
        if len(labels) < len(self.data) + 1:
            warnings.warn(
                (
                    f"Number of labels ({len(labels)}) does not match "
                    f"number of columns ({num_columns}). "
                    "filling missing labels with empty strings."
                )
            )
            labels += [""] * (num_columns - len(labels))
        headers = ["---" for _ in range(num_columns)]
        return (
            f"| {labels[0]} | "
            + " | ".join([labels[i] for i in range(1, num_columns)])
            + " |\n"
            + "| "
            + " | ".join(headers)
            + " |"
        )

    def _create_regular_table(self) -> str:
        """Create a regular (non-transposed) version of the table.

        Returns:
            str: Markdown formatted table.
        """
        headers = "| " + self._create_regular_headers() + " |"
        divider = "|" + "|".join([" --- " for _ in self.data[0].keys()]) + "|"
        table_rows = self._create_regular_table_rows()

        return headers + "\n" + divider + "\n" + table_rows

    def _create_regular_headers(self) -> str:
        """Create headers for the regular table.

        Returns:
            str: Table headers.
        """
        return " | ".join(self.data[0].keys())

    def _value_to_string(self, value: Any) -> str:
        """Convert the given value to a string. If it's a floating point number,
        it will be formatted according to the precision attribute.

        Args:
            value (Any): The value to be converted.

        Returns:
            str: The string representation of the value.
        """
        if isinstance(value, float) and self.precision is not None:
            return f"{value:.{self.precision}f}"
        return str(value)

    def _create_regular_table_rows(self) -> str:
        """Create table rows for the non-transposed table.

        Returns:
            str: Formatted table rows for the regular table.
        """
        rows = []
        for row in self.data:
            rows.append(
                "| "
                + " | ".join([self._value_to_string(value) for value in row.values()])
                + " |"
            )
        return "\n".join(rows)

    def _create_table_rows(self, data: Dict[str, List[Any]]) -> str:
        """Create table rows for the given data.

        Args:
            data (Dict[str, List[Any]]): Data to be used for row creation.

        Returns:
            str: Formatted table rows.
        """
        rows = []
        for key, values in data.items():
            rows.append(
                "| "
                + key
                + " | "
                + " | ".join([self._value_to_string(value) for value in values])
                + " |"
            )
        return "\n".join(rows)

    def __str__(self) -> str:
        return self.dict_to_md_table(transpose=self.transpose, precision=self.precision)



from ._base import MdElement


class MdHeader(MdElement):
    """Represents a Markdown header.

    Attributes:
        content (str): The content of the header.
        level (int): The header level.

    Examples:
        >>> from mdfy.elements import MdHeader
        >>>
        >>> header = MdHeader("This is a header")
        >>> print(header)
        # This is a header
        >>>
        >>> header = MdHeader("This is a header", level=2)
        >>> print(header)
        ## This is a header
    """

    def __init__(self, content: str, level: int = 1) -> None:
        """Initializes an instance of the MdHeader class to represent a Markdown header.

        Args:
            content (str): The content of the header.
            level (int, optional): The header level. Defaults to 1.
        """
        self.content = content
        self.level = level

    def __str__(self) -> str:
        """Returns a string representation of the header in Markdown format.

        Returns:
            str: String representation of the header.
        """
        return "#" * self.level + " " + self.content



from typing import List as TypedList
from typing import Union

from ._base import MdElement


class MdList(MdElement):
    """Represents a Markdown list.

    Attributes:
        items (List[Union[str, MdList]]): List of items, where each item can be a string or another MdList.
        depth (int): Current depth of the list.
        indent (int): indent size of the list.
        numbered (bool): If True, the list will be numbered. Otherwise, it will be bulleted.

    Examples:
        >>> from mdfy.elements import MdList
        >>>
        >>> list = MdList(["item 1", "item 2"])
        >>> print(list)
        - item 1
        - item 2
        >>>
        >>> list = MdList(["item 1", "item 2"], numbered=True)
        >>> print(list)
        1. item 1
        2. item 2
        >>>
        >>> list = MdList([
        ...     "item 1",
        ...     [
        ...         "item 2.1",
        ...         "item 2.2"
        ...     ]
        ... ])
        >>> print(list)
        - item 1
            - item 2.1
            - item 2.2
    """

    def __init__(
        self,
        items: TypedList[Union[str, "MdList"]],
        depth: int = 0,
        indent: int = 4,
        numbered: bool = False,
    ) -> None:
        """Initialize a MdList instance.

        Args:
            items (List[Union[str, MdList]]): List of items, where each item can be a string or another MdList.
            That's a [A:bold] statement (bool, optional): If True, the list will be numbered. Otherwise, it will be bulleted. Defaults to True.
            depth (int, optional): Current depth of the list. Defaults to 0.
            indent (int, optional): indent size of the list. Defaults to 4.

        """
        self.items = items
        self.numbered = numbered
        self.depth = depth
        self.indent = indent

    def _process_items(self, items: TypedList[Union[str, "MdList"]], depth: int) -> str:
        """Process the given list of items and convert them to a markdown string.

        Args:
            items (List[Union[str, MdList]]): List of items to be processed.
            depth (int): Current depth of the list.

        Returns:
            str: Formatted markdown string.
        """
        return "\n".join(f"{self._process_item(item, depth)}" for item in items)

    def _process_item(self, item: Union[str, "MdList"], depth: int) -> str:
        """Process an individual item. If the item is a list, process it recursively.

        Args:
            item (Union[str, MdList]): The item to be processed.
            depth (int): Current depth of the list.

        Returns:
            str: Formatted markdown string for the item.
        """
        if isinstance(item, list):
            # If the item is a list, process it recursively with one deeper level
            return self._process_items(item, depth + 1)

        prefix = " " * self.indent * depth
        marker = "1." if self.numbered else "-"
        return f"{prefix}{marker} {item}"

    def __str__(self) -> str:
        """Returns a string representation of the list in Markdown format.

        Returns:
            str: Formatted markdown string for the entire list.
        """
        return self._process_items(self.items, self.depth)



class MdElement:
    """Represents a Markdown element.

    Attributes:
        content (str or MdElement): The content of the element.
    """

    def __str__(self) -> str:
        raise NotImplementedError

    def to_str(self) -> str:
        """Returns a string representation of the element in Markdown format.

        Returns:
            str: String representation of the element.
        """

        return str(self)
